package br.com.rjconsultores.auditador.auditadores;

import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import br.com.rjconsultores.auditador.annotations.AuditarAtributo;
import br.com.rjconsultores.auditador.annotations.AuditarClasse;
import br.com.rjconsultores.auditador.annotations.AuditarEntidade;
import br.com.rjconsultores.auditador.annotations.AuditarID;
import br.com.rjconsultores.auditador.annotations.AuditarIDComposta;
import br.com.rjconsultores.auditador.annotations.AuditarLista;
import br.com.rjconsultores.auditador.annotations.AuditarMetodo;
import br.com.rjconsultores.auditador.annotations.NaoAuditar;
import br.com.rjconsultores.auditador.enums.AuditadorTipoAlteracao;
import br.com.rjconsultores.auditador.enums.TipoExtracaoDados;
import br.com.rjconsultores.auditador.exceptions.AuditadorException;
import br.com.rjconsultores.auditador.interfaces.Auditavel;
import br.com.rjconsultores.auditador.interfaces.AuditavelTelaAlternativa;
import br.com.rjconsultores.auditador.model.AuditadorObjects;

/**
 * Classe responsvel para listar as alteraes do objeto.
 * 
 * @author wilian
 *
 */
public class Auditador {
	
	AuditadorAtributo auditadorAtributo;
	AuditadorEntidade auditadorEntidade;
	AuditadorList auditadorList;
	ExtrairAnotacao extrairAnotacao;
	
	private AuditadorObjects padrao;
	private TipoExtracaoDados tipoExtracaoDados;
	
	private Auditador() {
		super();
	}
	
	public static Auditador getInstance( ) {
		Auditador auditar = new Auditador();
		auditar.auditadorAtributo = new AuditadorAtributo(auditar);
		auditar.auditadorEntidade = new AuditadorEntidade(auditar);
		auditar.auditadorList = new AuditadorList(auditar);
		auditar.extrairAnotacao = new ExtrairAnotacao();
		return auditar;
	}
	
	/**
	 * Responsvel por auditar os registros.
	 * 
	 * @param original
	 * @param novo
	 * @param empresaId
	 * @return
	 */
	public List<AuditadorObjects> auditar(Object original, Object novo) {
		List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
		try {
			
			if(isObjetosNulos(original, novo) || !iniciarObjetoPadrao(novo != null ? novo : original, original == null ? AuditadorTipoAlteracao.INCLUSAO : AuditadorTipoAlteracao.ALTERACAO)) {
				return lsRetorno;
			}
			
			AuditadorObjects objectNovo = auditarObjetoNovo(null, original, novo);
			if(objectNovo != null) {
				lsRetorno.add(objectNovo);
				return lsRetorno;
			}
			
			lsRetorno.addAll(auditarCampos(original, novo, padrao.getClasseAlterada()));
			return lsRetorno;
		} catch (Exception e) {
			throw new AuditadorException(e.getMessage(), e);
		} catch (Throwable e) {
			throw new AuditadorException(e.getMessage(), e);
		}
	}
	
	/**
	 * Responsvel por auditar os registros filhos do objeto de referencia.
	 * 
	 * @param objetoReferencia
	 * @param original
	 * @param novo
	 * @return
	 */
	public List<AuditadorObjects> auditar(Object objetoReferencia, Object original, Object novo) {
		List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
		try {
			
			if(!iniciarObjetoPadrao(objetoReferencia, AuditadorTipoAlteracao.ALTERACAO)) {
				return lsRetorno;
			}
			
			AuditadorObjects objectNovo = auditarObjetoNovo(null, original, novo);
			if(objectNovo != null) {
				lsRetorno.add(objectNovo);
				return lsRetorno;
			}
			
			lsRetorno.addAll(auditarCampos(original, novo, original != null ? original.getClass().getSimpleName() : novo.getClass().getSimpleName()));
			return lsRetorno;
		} catch (Exception e) {
			throw new AuditadorException(e.getMessage(), e);
		} catch (Throwable e) {
			throw new AuditadorException(e.getMessage(), e);
		}
	}
	
	/**
	 * Responsvel por auditar a lista quando a alterao acontece exclusivamente nos atributos lista do objeto
	 * 
	 * @param objeto
	 * @param nomeCampoListAuditar
	 * @param original
	 * @param novo
	 * @return
	 */
	public List<AuditadorObjects> auditarList(Object objeto, String nomeCampoListAuditar, List<?> original, List<?> novo) {
		List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
		try {
			if(!iniciarObjetoPadrao(objeto, AuditadorTipoAlteracao.ALTERACAO) || isObjetosNulos(original, novo)) {
				return lsRetorno;
			}
			
			Class<?> classe = objeto.getClass();
			
			List<AccessibleObject> camposAuditaveis = extrairCamposAuditaveis(classe);
			for (AccessibleObject campo : camposAuditaveis) {
				campo.setAccessible(true);
				String nome = extrairDeclaredNomeCampo(campo);

				if(!nome.equals(nomeCampoListAuditar)) {
					continue;
				}
				
				NaoAuditar naoAuditar = extrairAnotacao.extrairNaoAuditar(campo);
				if(naoAuditar != null) {
					continue;
				}
				
				List<Object> lsOriginal = new ArrayList<Object>();
				if(original != null) {
					lsOriginal.addAll(original);
				}
				
				List<Object> lsNovo = new ArrayList<Object>();
				if(novo != null) {
					lsNovo.addAll(novo);
				}
				
				AuditarLista auditarLista = extrairAnotacao.extrairAuditarLista(campo);
				lsRetorno.addAll(auditadorList.auditarList(campo, auditarLista != null ? auditarLista.auditarEntidades() : false, lsOriginal, lsNovo));
				
				break;
			}
		} catch (Exception e) {
			throw new AuditadorException(e.getMessage(), e);
		} catch (Throwable e) {
			throw new AuditadorException(e.getMessage(), e);
		}
		return lsRetorno;
	}
	
	/**
	 * Responsvel por auditar quando o objeto  excludo.
	 * 
	 * @param objeto
	 * @param empresaId 
	 * @return
	 */
	public List<AuditadorObjects> auditarExclusao(Object objeto) {
		try {
			List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
			
			if(isObjetosNulos(null, objeto) || !iniciarObjetoPadrao(objeto != null ? objeto : null, AuditadorTipoAlteracao.EXCLUSAO)) {
				return lsRetorno;
			}
			
			AuditadorObjects objetoExcluido = auditarObjetoExcluido(objeto);
			if(objetoExcluido != null) {
				lsRetorno.add(objetoExcluido);
			}
			
			return lsRetorno;
		} catch (Exception e) {
			throw new AuditadorException(e.getMessage(), e);
		} catch (Throwable e) {
			throw new AuditadorException(e.getMessage(), e);
		}
	}
	
	/**
	 * Responsvel por auditar quando o objeto sofre uma alterao personalizada.
	 * 
	 * @param objeto
	 * @param valorNovo
	 * @param empresaId 
	 * @return
	 */
	public List<AuditadorObjects> auditarPersonalizado(Object objeto, String valorNovo) {
		try {
			List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
			
			if(isObjetosNulos(null, objeto) || !iniciarObjetoPadrao(objeto != null ? objeto : null, AuditadorTipoAlteracao.ALTERACAO)) {
				return lsRetorno;
			}
			
			AuditadorObjects objetoAuditado = auditarObjetoPersonalizado(objeto, AuditadorTipoAlteracao.ALTERACAO, valorNovo);
			if(objetoAuditado != null) {
				lsRetorno.add(objetoAuditado);
			}
			
			return lsRetorno;
		} catch (Exception e) {
			throw new AuditadorException(e.getMessage(), e);
		} catch (Throwable e) {
			throw new AuditadorException(e.getMessage(), e);
		}
	}	
	
	protected AuditadorObjects auditarObjetoNovo(AccessibleObject campo, Object original, Object novo) throws CloneNotSupportedException {
		if(original == null && novo != null) {
			AuditadorObjects auditado = padrao.clonar();
			auditado.setValorNovo(String.format("Registro Criado [%s]", extrairTextoInclusaoExclusao(novo)));
			auditado.setCampoAlterado(extrairNomeCampo(campo));
			return auditado;
		}
		return null;
	}
	
	protected AuditadorObjects auditarObjetoPersonalizado(Object objeto, AuditadorTipoAlteracao tipoAlteracao, String valorNovo) throws CloneNotSupportedException {
		if(objeto != null) {
			AuditadorObjects auditado = padrao.clonar();
			auditado.setTipoAlteracao(tipoAlteracao != null ? tipoAlteracao : AuditadorTipoAlteracao.ALTERACAO);
			auditado.setValorNovo(String.format("%s [%s]", valorNovo, extrairTextoInclusaoExclusao(objeto)));
			return auditado;
		}
		return null;
	}
	
	protected AuditadorObjects auditarObjetoExcluido(Object objeto) throws CloneNotSupportedException {
		return auditarObjetoPersonalizado(objeto, AuditadorTipoAlteracao.EXCLUSAO, "Registro excluido");
	}
	
	protected AuditadorObjects auditarCampo(AccessibleObject campo, Object valorOriginal, Object valorNovo, String nomeClasse) throws CloneNotSupportedException {
		if(isDiferente(valorOriginal, valorNovo)) {
			String nomeCampo = extrairNomeCampo(campo);
			AuditadorObjects auditado = padrao.clonar();
			auditado.setTipoAlteracao(AuditadorTipoAlteracao.ALTERACAO);
			auditado.setCampoAlterado(nomeCampo);
			auditado.setValorAnterior(valorOriginal != null ? valorOriginal.toString() : null);
			auditado.setValorNovo(valorNovo != null ? valorNovo.toString() : null);
			auditado.setClasseAlterada(nomeClasse);
			return auditado;
		}
		return null;
	}

	protected String extrairNomeClasse(Object original, Object novo) {
		if(original != null) {
			return original.getClass().getSimpleName();
		}
		if(novo != null) {
			return novo.getClass().getSimpleName();
		}
		return null;
	}
	
	protected String extrairNomeCampo(AccessibleObject campo) {
		String nomeCampo = null;
		if(campo != null) {
			AuditarID auditarID = extrairAnotacao.extrairAuditarID(campo);
			if(auditarID != null && StringUtils.isNotBlank(auditarID.nome())) {
				nomeCampo = auditarID.nome();
			}
			AuditarAtributo auditarAtributo = extrairAnotacao.extrairAuditarAtributo(campo);
			if(auditarAtributo != null && StringUtils.isNotBlank(auditarAtributo.nome())) {
				nomeCampo = auditarAtributo.nome();
			}
			AuditarEntidade auditarEntidade = extrairAnotacao.extrairAuditarEntidade(campo);
			if(auditarEntidade != null && StringUtils.isNotBlank(auditarEntidade.nome())) {
				nomeCampo = auditarEntidade.nome();
			}
			AuditarLista auditarLista = extrairAnotacao.extrairAuditarLista(campo);
			if(auditarLista != null && StringUtils.isNotBlank(auditarLista.nome())) {
				nomeCampo = auditarLista.nome();
			}
			AuditarMetodo auditarMetodo = extrairAnotacao.extrairAuditarMetodo(campo);
			if(auditarMetodo != null && StringUtils.isNotBlank(auditarMetodo.nome())) {
				nomeCampo = auditarMetodo.nome();
			}
			
			if(StringUtils.isBlank(nomeCampo)) {
				nomeCampo = extrairDeclaredNomeCampo(campo);
			}
		}
		return nomeCampo;
	}

	protected String extrairDeclaredNomeCampo(AccessibleObject campo) {
		if(isExtrairDadosAtributo()) {
			return ((Field) campo).getName();
		} if(isExtrairDadosMetodo()) {
			return ((Method) campo).getName();
		}
		return null;
	}

	protected boolean isDiferente(Object valorOriginal, Object valorNovo) {
		if(valorOriginal == null && valorNovo == null) {
			return false;
		}
		
		if(valorOriginal != null && valorNovo == null) {
			return true;
		}
		
		if(valorOriginal == null && valorNovo != null) {
			return true;
		}
		
		return !valorOriginal.equals(valorNovo);
	}

	protected List<AuditadorObjects> auditarCampos(Object original, Object novo, String nomeClasse) throws IllegalArgumentException, IllegalAccessException {
		List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
		if(!isObjetosNulos(original, novo)) {
			Class<?> classe = original != null ? original.getClass() : novo.getClass();
			List<AccessibleObject> camposAuditaveis = extrairCamposAuditaveis(classe);
			for (AccessibleObject campo : camposAuditaveis) {
				AuditarEntidade auditarEntidade = extrairAnotacao.extrairAuditarEntidade(campo);
				AuditarLista auditarLista = extrairAnotacao.extrairAuditarLista(campo);
				
				if(auditarEntidade != null) {
					lsRetorno.addAll(auditadorEntidade.auditar(campo, original, novo, null));
				} else if(auditarLista != null) {
					lsRetorno.addAll(auditadorList.auditar(campo, original, novo, auditarLista.auditarEntidades()));
				} else {
					lsRetorno.addAll(auditadorAtributo.auditar(campo, original, novo, nomeClasse));
				}
			}
		}
		return lsRetorno;
		
	}

	protected List<AccessibleObject> extrairCamposAuditaveis(Class<?> classe) {
		List<AccessibleObject> campo = new ArrayList<AccessibleObject>();
		if(isExtrairDadosAtributo()) {
			campo = extrairFields(classe);
		} else if(isExtrairDadosMetodo()) {
			campo = extrairMethods(classe);
		}
		return campo;
	}

	protected List<AccessibleObject> extrairMethods(Class<?> classe) {
		List<AccessibleObject> campos = new ArrayList<AccessibleObject>();
		Method[] methods = classe.getDeclaredMethods();
		for (Method method : methods) {
			method.setAccessible(true);
			
			NaoAuditar naoAuditar = extrairAnotacao.extrairNaoAuditar(method);
			if(naoAuditar != null) {
				continue;
			}
			
			if(!method.getReturnType().equals(Void.TYPE) && method.getParameterTypes().length == 0  &&
			   !isMethodIgnorado(method)) {
				campos.add(method);
			}
		}
		return campos;
	}

	private boolean isMethodIgnorado(Method method) {
		return method.getName().equalsIgnoreCase("clone") ||
			   method.getName().equalsIgnoreCase("clonar") ||
			   method.getName().equalsIgnoreCase("getCloneObject") ||
			   method.getName().equalsIgnoreCase("hashCode") ||
			   method.getName().equalsIgnoreCase("equals");
	}

	protected List<AccessibleObject> extrairFields(Class<?> classe) {
		List<AccessibleObject> campos = new ArrayList<AccessibleObject>();
		Field[] fields = classe.getDeclaredFields();
		for (Field field : fields) {
			field.setAccessible(true);
			
			NaoAuditar naoAuditar = extrairAnotacao.extrairNaoAuditar(field);
			if(naoAuditar != null) {
				continue;
			}
			
			campos.add(field);
		}
		return campos;
	}

	protected Object getValor(AccessibleObject campo, Object object) throws IllegalArgumentException, IllegalAccessException, ParseException, InvocationTargetException {
		return getValor(campo, object, null);
	}
	
	protected Object getValor(AccessibleObject campo, Object object, String pattern) throws IllegalArgumentException, IllegalAccessException, ParseException, InvocationTargetException {
		if(campo == null || object == null) {
			return null;
		}
		Object retorno = extrairValor(campo, object);
		
		Object atributo = null;
		if (retorno instanceof Date) {
			Date aux = (Date) retorno;

			atributo = getStringDate(aux, pattern);
		} else if (retorno instanceof Integer || retorno instanceof Long ) {
			atributo = retorno.toString();
		} else if (retorno instanceof Number) {
			atributo = getValorFormatado(retorno.toString());
		} else if(retorno instanceof String) {
			atributo = StringUtils.isNotBlank(retorno.toString()) ? retorno.toString() : "nulo";
		} else {
			return retorno;
		}
		return atributo;
		
	}
	
	protected String getValorFormatado(String value) throws ParseException {
		
		if(value != null && StringUtils.isNotBlank(value)) {
			NumberFormat nf = NumberFormat.getInstance(new Locale("pt", "br"));
			BigDecimal valor =  new BigDecimal(nf.parse(value.replace(".", ",")).toString());
			
			return nf.format(valor.doubleValue());
		}
		
		return null;
	}
	
	protected String getStringDate(java.util.Date d, String pattern) {
		if(StringUtils.isBlank(pattern)) {
			pattern = "dd/MM/yyyy HH:mm";
		}
		
		SimpleDateFormat sdf = new SimpleDateFormat(pattern);
		if (d != null) {
			return sdf.format(d);
		}
		return null;
	}	
	
	@SuppressWarnings("unchecked")
	protected List<Object> getValorList(AccessibleObject campo, Object object) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		if(campo == null || object == null) {
			return new ArrayList<Object>();
		}
		Object aux = extrairValor(campo, object);
		
		List<Object> retorno = new ArrayList<Object>();
		if(aux != null && aux instanceof Collection) {
			retorno.addAll((Collection<Object>) aux);
		}
		
		return retorno;
	}

	protected boolean isObjetosNulos(Object original, Object novo) {
		return novo == null && original == null;
	}
	
	protected boolean isObjetoNovo(Object original, Object novo) {
		return novo != null && original == null;
	}

	protected boolean iniciarObjetoPadrao(Object objeto, AuditadorTipoAlteracao tipoAlteracao) throws IllegalArgumentException, IllegalAccessException, SecurityException, NoSuchFieldException, ParseException, NoSuchMethodException, InvocationTargetException {
		padrao = null;
		Class<?> classe = objeto.getClass();
		AuditarClasse auditarClasse = extrairAnotacao.extrair(classe);
		if(auditarClasse != null) {
			tipoExtracaoDados = auditarClasse.tipoExtracaoDados();
			padrao = new AuditadorObjects();
			padrao.setTipoAlteracao(tipoAlteracao);
			padrao.setTela(getNomeTela(auditarClasse, objeto));
			padrao.setIdAuditado(getIdEntidade(objeto));
			padrao.setClassePrincipal(StringUtils.isNotBlank(auditarClasse.nome()) ? auditarClasse.nome() : classe.getSimpleName());
			
		}
		return padrao != null;
	}

	protected String getNomeTela(AuditarClasse auditarClasse, Object objeto) {
		if(objeto instanceof AuditavelTelaAlternativa) {
			AuditavelTelaAlternativa aux = (AuditavelTelaAlternativa) objeto;
			if(aux != null && StringUtils.isNotBlank(aux.getTelaAlternativa())) {
				return aux.getTelaAlternativa();
			}
		}
		return auditarClasse.tela();
	}

	protected String getIdEntidade(Object objeto) throws IllegalArgumentException, IllegalAccessException, ParseException, InvocationTargetException {
		Object valor = null;
		Class<?> classe = objeto.getClass();
		List<AccessibleObject> camposAuditaveis = extrairCamposAuditaveis(classe);
		for (AccessibleObject campo : camposAuditaveis) {
			campo.setAccessible(true);
			AuditarID auditarID = extrairAnotacao.extrairAuditarID(campo);
			if(auditarID != null) {
				valor = getValor(campo, objeto, auditarID.pattern());
			}
			
			if(valor == null) {
				AuditarIDComposta auditarIDComposta = extrairAnotacao.extrairAuditarIDComposta(campo);
				if(auditarIDComposta != null) {
					valor = getIdEntidadeComposta(campo, objeto);
				}
			}
			
		}
		return valor != null ? valor.toString() : null;
	}

	protected Object getIdEntidadeComposta(AccessibleObject campo, Object objeto) throws IllegalArgumentException, IllegalAccessException, ParseException, InvocationTargetException {
		StringBuilder sb = new StringBuilder();
		if(campo != null && objeto != null) {
			Object valorID = extrairValor(campo, objeto);
			if(valorID != null) {
				Class<?> classe = valorID.getClass();
				List<AccessibleObject> camposAuditaveis = extrairCamposAuditaveis(classe);
				for (AccessibleObject campoId : camposAuditaveis) {
					campoId.setAccessible(true);
					AuditarID auditarID = extrairAnotacao.extrairAuditarID(campoId);
					if(auditarID != null) {
						Object valor = getValor(campoId, valorID, auditarID.pattern());
						if(valor != null) {
							if(sb.length() > 0) {
								sb.append("-");
							}
							sb.append(valor.toString());
						}
					}
				}
			}
		}
		return sb.length() > 0 ? sb.toString() : null;
	}

	protected AuditadorObjects getPadrao() {
		try {
			return padrao.clonar();
		} catch (Exception e) {
			throw new AuditadorException("No foi possvel clonar o objeto padro", e);
		}
	}

	protected String extrairTextoInclusaoExclusao(Object objeto) {
		if(objeto instanceof Auditavel) {
			Auditavel<?> auditavelAux = (Auditavel<?>) objeto;
			return auditavelAux.getTextoInclusaoExclusao();
		}
		return "";
	}
	
	protected Object extrairValor(AccessibleObject campo, Object object) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		if(isExtrairDadosAtributo()) {
			return ((Field) campo).get(object);
		} else if(isExtrairDadosMetodo()) {
			return ((Method) campo).invoke(object);
		}
		return null;
	}
	
	protected boolean isExtrairDadosAtributo() {
		return TipoExtracaoDados.ATRIBUTO.equals(tipoExtracaoDados);
	}
	
	protected boolean isExtrairDadosMetodo() {
		return TipoExtracaoDados.METODO.equals(tipoExtracaoDados);
	}

	protected String extrairPattern(AccessibleObject campoEntidade) {
		String pattern = null;
		if(isExtrairDadosAtributo()) {
			AuditarAtributo auditarAtributo =  extrairAnotacao.extrairAuditarAtributo(campoEntidade);
			if(auditarAtributo != null) {
				pattern = auditarAtributo.pattern();
			}
		} else if(isExtrairDadosMetodo()) {
			AuditarMetodo auditarMetodo =  extrairAnotacao.extrairAuditarMetodo(campoEntidade);
			if(auditarMetodo != null) {
				pattern = auditarMetodo.pattern();
			}
		}
		return pattern;
	}
	
	protected AccessibleObject extrairAccessibleObject(Object objeto, String nome) throws SecurityException, NoSuchFieldException, NoSuchMethodException {
		if(objeto != null && StringUtils.isNotBlank(nome)) {
			Class<?> classe = objeto.getClass();
			if(isExtrairDadosAtributo()) {
				Field field = classe.getDeclaredField(nome);
				return field;
			} else if(isExtrairDadosMetodo()) {
				Method method = classe.getDeclaredMethod(nome);
				return method;
			}
			
		}
		return null;
	}

}
