package br.com.rjconsultores.auditador.auditadores;

import java.lang.reflect.AccessibleObject;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import br.com.rjconsultores.auditador.enums.AuditadorTipoAlteracao;
import br.com.rjconsultores.auditador.exceptions.AuditadorException;
import br.com.rjconsultores.auditador.model.AuditadorObjects;

class AuditadorList {

	protected Auditador auditador;

	public AuditadorList(Auditador auditar) {
		super();
		this.auditador = auditar;
	}
	
	public List<AuditadorObjects> auditar(AccessibleObject campo, Object original, Object novo, boolean auditarEntidades) {
		List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
		try {
			
			List<Object> lsOriginal = auditador.getValorList(campo, original);
			List<Object> lsNovo = auditador.getValorList(campo, novo);
			
			lsRetorno.addAll(auditarList(campo, auditarEntidades, lsOriginal, lsNovo));
		} catch (Exception e) {
			throw new AuditadorException(e.getMessage(), e);
		} catch (Throwable e) {
			throw new AuditadorException(e.getMessage(), e);
		}
		return lsRetorno;
	}

	public List<AuditadorObjects> auditarList(AccessibleObject campo, boolean auditarEntidades,
			List<Object> lsOriginal, List<Object> lsNovo) throws CloneNotSupportedException {
		List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
		
		List<Object> lsObjetosInseridos = new ArrayList<Object>(0);
		List<Object> lsObjetosRemovidos = new ArrayList<Object>(0);
		List<Object> lsObjetosMantidos = new ArrayList<Object>(0);
		
		if(lsNovo != null) {
			lsObjetosInseridos.addAll(lsNovo);
			if(lsOriginal != null) {
				lsObjetosInseridos.removeAll(lsOriginal);
			}
			lsRetorno.addAll(auditarRegistros(campo, lsObjetosInseridos, true));
			
			if(auditarEntidades) {
				lsObjetosMantidos.addAll(lsNovo);
			}
		}
		
		if(lsOriginal != null) {
			lsObjetosRemovidos.addAll(lsOriginal);
			if(lsNovo != null) {
				lsObjetosRemovidos.removeAll(lsNovo);
			}
			lsRetorno.addAll(auditarRegistros(campo, lsObjetosRemovidos, false));
			
			if(auditarEntidades) {
				lsObjetosMantidos.retainAll(lsOriginal);
			}
		}
		
		if(auditarEntidades) {
			for (Object object : lsObjetosMantidos) {
				Object originalEntidade = lsOriginal.get(lsOriginal.indexOf(object));
				Object novoEntidade = lsNovo.get(lsNovo.indexOf(object));
				String nomeClasse = auditador.extrairNomeClasse(originalEntidade, novoEntidade);
				
				lsRetorno.addAll(auditador.auditadorEntidade.auditar(null, originalEntidade, novoEntidade, nomeClasse));
			}
		}
		return lsRetorno;
	}

	protected List<AuditadorObjects> auditarRegistros(AccessibleObject campo, List<Object> lsObjetosRemovidos, boolean inserido)
			throws CloneNotSupportedException {
		List<AuditadorObjects> lsRetorno = new ArrayList<AuditadorObjects>();
		String nomeCampo = auditador.extrairNomeCampo(campo);
		for (Object removido : lsObjetosRemovidos) {
			AuditadorObjects objectInserido = null;

			if(inserido) {
				objectInserido = auditarObjetoInserido(nomeCampo, removido);
			} else {
				objectInserido = auditarObjetoRemovido(nomeCampo, removido);
			}
			
			if(objectInserido != null) {
				lsRetorno.add(objectInserido);
			}
		}
		return lsRetorno;
	}

	protected boolean isObjetosList(Object objeto) {
		return objeto != null && objeto instanceof Collection;
	}
	
	protected AuditadorObjects auditarObjetoInserido(String nomeCampo, Object objeto) throws CloneNotSupportedException {
		if(objeto != null) {
			AuditadorObjects auditado = criarObjetoInseridoRemovido(nomeCampo, objeto);
			auditado.setValorNovo(String.format("Registro inserido [%s]", auditador.extrairTextoInclusaoExclusao(objeto)));
			return auditado;
		}
		return null;
	}
	
	private AuditadorObjects criarObjetoInseridoRemovido(String nomeCampo, Object objeto) throws CloneNotSupportedException {
		AuditadorObjects auditado = auditador.getPadrao().clonar();
		auditado.setTipoAlteracao(AuditadorTipoAlteracao.ALTERACAO);
		auditado.setCampoAlterado(nomeCampo);
		auditado.setClasseAlterada(objeto.getClass().getSimpleName());
		return auditado;
	}
	
	protected AuditadorObjects auditarObjetoRemovido(String nomeCampo, Object objeto) throws CloneNotSupportedException {
		if(objeto != null) {
			AuditadorObjects auditado = criarObjetoInseridoRemovido(nomeCampo, objeto);
			auditado.setValorNovo(String.format("Registro removido [%s]", auditador.extrairTextoInclusaoExclusao(objeto)));
			return auditado;
		}
		return null;
	}

}
